/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * This file is part of the PSIQS package version 01 from 2016-01-07, 
 * developed by Tilman Neumann, Edemissen, Germany.
 * 
 * Use and redistribution of PSIQS are free for non-commercial purposes, as long as this
 * header is not modified or removed. Any commercial usage of the whole or parts
 * of PSIQS requires the written consent of the author.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

package de.tilman_neumann.math.factor;

import java.math.BigInteger;
import java.util.Map;

import org.apache.log4j.Logger;

import de.tilman_neumann.math.base.bigint.primes.ProbablePrimeTest;
import de.tilman_neumann.types.SortedMultiset;
import de.tilman_neumann.types.SortedMultiset_TreeMapImpl;

import static de.tilman_neumann.math.base.bigint.BigIntConstants.*;

/**
 * Abstraction of integer factorization algorithms.
 * This class provides a framework to find the complete prime factorization of N,
 * requiring only to implement the method findSingleFactor(BigInteger).
 * 
 * @author Tilman Neumann
 */
abstract public class FactorAlgorithmBase implements SingleFactorFinder {
	@SuppressWarnings("unused")
	private static final Logger LOG = Logger.getLogger(FactorAlgorithmBase.class);

	private ProbablePrimeTest probablePrimeTest;

	/**
	 * Complete constructor.
	 * 
	 * @param primeTestBits log of probability of predicting false primes
	 */
	public FactorAlgorithmBase(int primeTestBits) {
		probablePrimeTest = new ProbablePrimeTest(primeTestBits);
	}
	
	/**
	 * Factoring of composite integers.
	 * 
	 * @param N Number to decompose into prime factors.
	 * @return Factors or null if N prime.
	 */
	public SortedMultiset<BigInteger> factor(BigInteger N) {
		SortedMultiset<BigInteger> factors = new SortedMultiset_TreeMapImpl<BigInteger>();
		// first get rid of case |N|<=1:
		if (N.abs().compareTo(ONE)<=0) {
			factors.add(N);
			return factors;
		}
		// make N positive:
		if (N.signum()<0) {
			factors.add(MINUS_ONE);
			N = N.abs();
		}
		// Remove multiples of 2:
		int lsb = N.getLowestSetBit();
		if (lsb > 0) {
			factors.add(TWO, lsb);
			N = N.shiftRight(lsb);
		}
		if (N.equals(ONE)) {
			// N was just a power of 2
			return factors;
		}
		
		// N contains other factors...
		SortedMultiset<BigInteger> otherFactors = factor_recurrent(N);
		//LOG.debug(this.factorAlg + ": pow2Factors=" + factors + ", otherFactors=" + otherFactors);
		factors.addAll(otherFactors);
		//LOG.debug(this.factorAlg + ": => all factors = " + factors);
		return factors;
	}
	
	private SortedMultiset<BigInteger> factor_recurrent(BigInteger N) {
		SortedMultiset<BigInteger> factors = new SortedMultiset_TreeMapImpl<BigInteger>();
		if (probablePrimeTest.isProbablePrime(N)) {
			// N is probably prime. In exceptional cases this prediction
			// may be wrong and N composed -> then we would falsely predict N to be prime.
			//LOG.debug(N + " is probable prime.");
			factors.add(N);
			return factors;
		} // else: N is surely not prime

		// this is the expensive part:
		// find a factor of N, where N is composed and has no 2s
		BigInteger factor1 = findSingleFactor(N);
		// Is it possible to further decompose the parts?
		//LOG.debug("found factor1 = " + factor1, new Throwable());
		SortedMultiset<BigInteger> subfactors1 = factor_recurrent(factor1);
		//LOG.debug("subfactors of " + factor1 + " = " + subfactors1);
		factors.addAll(subfactors1);
		BigInteger factor2 = N.divide(factor1);
		SortedMultiset<BigInteger> subfactors2 = factor_recurrent(factor2);
		//LOG.debug("subfactors of " + factor2 + " = " + subfactors2);
		factors.addAll(subfactors2);
		//LOG.debug("result = " + factors);
		return factors;
	}
	
	/**
	 * Returns a product-like representation of the given factorization,
	 * with distinct keys separated by "*" and the multiplicity indicated by "^".
	 */
	public String getPrettyFactorString(SortedMultiset<BigInteger> factorization) {
		if (factorization.size()>0) {
			// Implementation note: Is faster with String than with StringBuffer!
			String factorStr = "";
			for (Map.Entry<BigInteger, Integer> entry : factorization.entrySet()) {
				factorStr += entry.getKey();
				Integer multiplicity = entry.getValue();
				if (multiplicity.intValue() > 1) {
					factorStr += "^" + multiplicity;
				}
				factorStr += " * ";
			}
			// remove the last ", "
			return factorStr.substring(0, factorStr.length()-3);
		}
		
		// no elements
		return "1";
	}
}
