/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * This file is part of the PSIQS package version 01 from 2016-01-07, 
 * developed by Tilman Neumann, Edemissen, Germany.
 * 
 * Use and redistribution of PSIQS are free for non-commercial purposes, as long as this
 * header is not modified or removed. Any commercial usage of the whole or parts
 * of PSIQS requires the written consent of the author.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

package de.tilman_neumann.math.factor.qs;

import java.math.BigInteger;

import org.apache.log4j.Logger;

import static de.tilman_neumann.math.base.bigint.BigIntConstants.*;

/**
 * Polynomial Q(x) = A(x)^2 - kN, A(x) = x+b, b=ceil(sqrt(kN)).
 * @author Tilman Neumann
 */
public class BasicQSPoly01 implements Poly {
	@SuppressWarnings("unused")
	private static final Logger LOG = Logger.getLogger(BasicQSPoly01.class);

	private BigInteger kN, ceil_sqrt_kN;
	
	// prime base
	private boolean hasNewPrimeBase;
	private int[] primesArray;
	private BigInteger[] primesArray_big;
	private byte[] logPArray;
	private float logBase;

	private int[] x1Array;
	private int[] x2Array;

	public BasicQSPoly01(BigInteger N, BigInteger kN, BigInteger ceil_sqrt_kN, boolean hasNewPrimeBase, int primeBaseSize, 
					int[] primesArray, BigInteger[] primesArray_big, byte[] logPArray, float logBase, int[] x1Array, int[] x2Array) {
		
		this.kN = kN;
		this.ceil_sqrt_kN = ceil_sqrt_kN;
		this.hasNewPrimeBase = hasNewPrimeBase;
		this.primesArray = primesArray;
		this.primesArray_big = primesArray_big;
		this.logPArray = logPArray;
		this.logBase = logBase;
		this.x1Array = x1Array;
		this.x2Array = x2Array;
	}

	@Override
	public BigInteger getkN() {
		return kN;
	}

	@Override
	public BigInteger getAParameter() {
		return ONE;
	}

	@Override
	public boolean hasNewPrimeBase() {
		return hasNewPrimeBase;
	}
	
	@Override
	public int[] getPrimesArray() {
		return primesArray;
	}

	@Override
	public BigInteger[] getPrimesArray_big() {
		return primesArray_big;
	}

	@Override
	public byte[] getLogPArray() {
		return logPArray;
	}

	@Override
	public float getLogBase() {
		return logBase;
	}

	@Override
	public int[] getX1Array() {
		return x1Array;
	}

	@Override
	public int[] getX2Array() {
		return x2Array;
	}

	@Override
	public BigInteger[] computeAandQ(int x) {
		BigInteger A = BigInteger.valueOf(x).add(ceil_sqrt_kN);  // A(x) = x+b, b=ceil(sqrt(kN))
		return new BigInteger[] {A, A.multiply(A).subtract(kN)}; // Q(x) = A(x)^2 - kN
	}
	
	/**
	 * @return all prime base elements that divide any Q(x).
	 * In SIQS, these are the q_l whose product gives the a-parameter.
	 * MPQS does not have such elements and returns null.
	 */
	@Override
	public int[] getPrimeBaseElementsDividingAnyQ() {
		// in basic QS there are no divisors that divide all Q(x)
		return null;
	}
	
	/**
	 * @return all primes not in the prime base that divide any Q(x).
	 * SIQS does not have such elements and returns null.
	 * MPQS has one such element, the q such that a=q^2.
	 */
	@Override
	public BigInteger[] getNonPrimeBaseElementsDividingAnyQ() {
		// in basic QS there are no divisors that divide all Q(x)
		return null;
	}
}
