/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * This file is part of the PSIQS package version 01 from 2016-01-07, 
 * developed by Tilman Neumann, Edemissen, Germany.
 * 
 * Use and redistribution of PSIQS are free for non-commercial purposes, as long as this
 * header is not modified or removed. Any commercial usage of the whole or parts
 * of PSIQS requires the written consent of the author.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

package de.tilman_neumann.math.factor.squfof;

import java.math.BigInteger;

import org.apache.log4j.Logger;

import de.tilman_neumann.math.base.bigint.SqrtExact;
import de.tilman_neumann.math.base.bigint.SqrtInt;
import de.tilman_neumann.math.base.bigint.sequence.IntegerSequence;
import de.tilman_neumann.math.base.bigint.sequence.SquarefreeSequence;
import de.tilman_neumann.math.factor.FactorAlgorithmBase;

import static de.tilman_neumann.math.base.bigint.BigIntConstants.*;

/**
 * Shanks' SQUFOF algorithm.<br/>
 * Implemented according to <link>http://en.wikipedia.org/wiki/Shanks'_square_forms_factorization</link>.
 * 
 * Final choice.
 * Stopping criterion: after a maximum number of iterations.
 * 
 * @author Tilman Neumann
 */
public class SquFoF extends FactorAlgorithmBase {
	@SuppressWarnings("unused")
	private static final Logger LOG = Logger.getLogger(SquFoF.class);

	// input
	private BigInteger N, kN, floor_sqrt_kN;

	// maximum number of iterations
	private long maxI;
	
	/**
	 * Standard constructor.
	 * @param kSequence multiplier generator
	 */
	public SquFoF(int primeTestBits) {
		super(primeTestBits);
	}

	@Override
	public String getName() {
		return "SquFoF";
	}
	
	/**
	 * Test the current N.
	 * @return factor, or null if no factor was found.
	 */
	public BigInteger findSingleFactor(BigInteger N) {
		this.N = N;
		// max iterations: there is no need to account for k, because expansions of smooth kN are typically not longer than those for N.
		// here we use the best choice found by experiments.
		this.maxI = (long) (1.5F*Math.pow(N.doubleValue(), 0.2));
		// compute best multiplier, determined by experiment
		int multiplier;
		switch (N.bitLength()) {
		case 20: case 21: case 22: /* and below */		multiplier = 240240;break; // tested with NCOUNT=100000
		case 48: case 49: case 50: 						multiplier = 240; 	break; // tested with NCOUNT=10000
		case 51: 										multiplier = 120;	break; // tested with NCOUNT=10000
		case 52: case 53: 								multiplier = 48; 	break; // tested with NCOUNT= 5000
		case 54: case 55: case 56: case 57: case 58: 	multiplier = 24; 	break; // tested with NCOUNT= 5000, 2500
		// case 59: 840 may be better than 1680 (tested with NCOUNT=1000)
		// case 60: 720 may be better than 1680 (tested with NCOUNT=1000)
		default: 										multiplier = 1680;
		}
		// sequence: for each new N start again with the first k
		IntegerSequence<BigInteger> kSequence = new SquarefreeSequence(BigInteger.valueOf(multiplier));
		kSequence.reset(N);
		while (true) {
			// get a new k, return immediately if kN is square
			this.kN = kSequence.next().multiply(N);
			BigInteger[] iSqrt = SqrtInt.iSqrt(kN);
			this.floor_sqrt_kN = iSqrt[0];
			if (floor_sqrt_kN.equals(iSqrt[1])) return N.gcd(floor_sqrt_kN);
			// search square Q_i
			BigInteger factor = test();
			if (factor!=null) return factor;
		}
	}

	protected BigInteger test() {
		// initialization for first iteration step
		long i = 0;
		BigInteger P_im1 = ONE;
		BigInteger P_i = floor_sqrt_kN;
		BigInteger Q_i = ONE;
		BigInteger Q_ip1 = kN.subtract(P_i.multiply(P_i)); // faster than P_i.pow(2)
		
		// first iteration step
		while (true) {
			// [McMath 2004] points out that we have to look for a square Q_i at some even i.
			// Here I test Q_i+1, so I have to look for square Q_i+1 at odd i!
			if (i%2==1) {
				BigInteger Q_ip1_sqrt = SqrtExact.exactSqrt(Q_ip1);
				if (Q_ip1_sqrt!=null) {
					// Q_i+1 is square -> do reverse iteration
					BigInteger factor = reverseIteration(P_i, Q_ip1_sqrt);
					if (factor!=null) return factor; // if factor is null we try another k
				}
			}
		
			// exit ?
			if (++i==maxI) return null;
			
			// keep values from last round
			P_im1 = P_i;
			BigInteger Q_im1 = Q_i;
			Q_i = Q_ip1;
			// compute next values
			BigInteger b_i = floor_sqrt_kN.add(P_im1).divideAndRemainder(Q_i)[0]; // floor(rational result)
			P_i = b_i.multiply(Q_i).subtract(P_im1);
			Q_ip1 = Q_im1.add(b_i.multiply(P_im1.subtract(P_i)));
		}
	}
	
	private BigInteger reverseIteration(BigInteger found_P, BigInteger found_Q_sqrt) {
		// initialization for second iteration step
		long i = 0;
		BigInteger b_i = floor_sqrt_kN.subtract(found_P).divideAndRemainder(found_Q_sqrt)[0]; // floor(rational result)
		BigInteger P_i = b_i.multiply(found_Q_sqrt).add(found_P);
		BigInteger Q_i = found_Q_sqrt;
		BigInteger Q_ip1 = kN.subtract(P_i.multiply(P_i)).divide(found_Q_sqrt); // faster than P_i.pow(2)
		
		// second iteration step
		BigInteger P_im1, Q_im1;
		do {
			// exit ?
			if (++i==maxI) return null;
			// keep values from last round
			P_im1 = P_i;
			Q_im1 = Q_i;
			Q_i = Q_ip1;
			// compute next values
			b_i = floor_sqrt_kN.add(P_im1).divideAndRemainder(Q_i)[0]; // floor(rational result)
			P_i = b_i.multiply(Q_i).subtract(P_im1);
			Q_ip1 = Q_im1.add(b_i.multiply(P_im1.subtract(P_i)));
		} while(!P_i.equals(P_im1));
		
		// result
		BigInteger gcd = N.gcd(P_i);
		return (gcd.compareTo(ONE)>0 && gcd.compareTo(N)<0) ? gcd : null;
	}
}
